"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = __importDefault(require("path"));
const ws_1 = __importDefault(require("ws"));
const hash_sum_1 = __importDefault(require("hash-sum"));
const chokidar_1 = __importDefault(require("chokidar"));
const vue_1 = require("./vue");
const fs_1 = require("fs");
const hmrClientPath = path_1.default.resolve(__dirname, '../../client/client.js');
exports.hmrMiddleware = ({ cwd, app, server }) => {
    app.use((ctx, next) => {
        if (ctx.path !== '/__hmrClient') {
            return next();
        }
        ctx.type = 'js';
        ctx.body = fs_1.createReadStream(hmrClientPath);
    });
    // start a websocket server to send hmr notifications to the client
    const wss = new ws_1.default.Server({ server });
    const sockets = new Set();
    wss.on('connection', (socket) => {
        sockets.add(socket);
        socket.send(JSON.stringify({ type: 'connected' }));
        socket.on('close', () => {
            sockets.delete(socket);
        });
    });
    wss.on('error', (e) => {
        if (e.code !== 'EADDRINUSE') {
            console.error(e);
        }
    });
    const notify = (payload) => sockets.forEach((s) => s.send(JSON.stringify(payload)));
    const watcher = chokidar_1.default.watch(cwd, {
        ignored: [/node_modules/]
    });
    watcher.on('change', async (file) => {
        const resourcePath = '/' + path_1.default.relative(cwd, file);
        const send = (payload) => {
            console.log(`[hmr] ${JSON.stringify(payload)}`);
            notify(payload);
        };
        if (file.endsWith('.vue')) {
            // check which part of the file changed
            const [descriptor, prevDescriptor] = await vue_1.parseSFC(cwd, file);
            if (!descriptor || !prevDescriptor) {
                // the file has never been accessed yet
                return;
            }
            if (!isEqual(descriptor.script, prevDescriptor.script)) {
                send({
                    type: 'reload',
                    path: resourcePath
                });
                return;
            }
            if (!isEqual(descriptor.template, prevDescriptor.template)) {
                send({
                    type: 'rerender',
                    path: resourcePath
                });
                return;
            }
            const prevStyles = prevDescriptor.styles || [];
            const nextStyles = descriptor.styles || [];
            if (prevStyles.some((s) => s.scoped) !== nextStyles.some((s) => s.scoped)) {
                send({
                    type: 'reload',
                    path: resourcePath
                });
            }
            nextStyles.forEach((_, i) => {
                if (!prevStyles[i] || !isEqual(prevStyles[i], nextStyles[i])) {
                    send({
                        type: 'style-update',
                        path: resourcePath,
                        index: i
                    });
                }
            });
            prevStyles.slice(nextStyles.length).forEach((_, i) => {
                send({
                    type: 'style-remove',
                    path: resourcePath,
                    id: `${hash_sum_1.default(resourcePath)}-${i + nextStyles.length}`
                });
            });
        }
        else {
            send({
                type: 'full-reload'
            });
        }
    });
};
function isEqual(a, b) {
    if (!a || !b)
        return false;
    if (a.content !== b.content)
        return false;
    const keysA = Object.keys(a.attrs);
    const keysB = Object.keys(b.attrs);
    if (keysA.length !== keysB.length) {
        return false;
    }
    return keysA.every((key) => a.attrs[key] === b.attrs[key]);
}
