"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const path_1 = __importDefault(require("path"));
const fs_1 = require("fs");
const resolveVue_1 = require("../resolveVue");
const hash_sum_1 = __importDefault(require("hash-sum"));
exports.vueMiddleware = ({ cwd, app }) => {
    app.use(async (ctx, next) => {
        if (!ctx.path.endsWith('.vue')) {
            return next();
        }
        const pathname = ctx.path;
        const query = ctx.query;
        const filename = path_1.default.join(cwd, pathname.slice(1));
        const [descriptor] = await parseSFC(cwd, filename, true /* save last accessed descriptor on the client */);
        if (!descriptor) {
            ctx.status = 404;
            return;
        }
        ctx.type = 'js';
        if (!query.type) {
            ctx.body = compileSFCMain(descriptor, pathname, query.t);
            return;
        }
        if (query.type === 'template') {
            ctx.body = compileSFCTemplate(cwd, descriptor.template, filename, pathname, descriptor.styles.some((s) => s.scoped));
            return;
        }
        if (query.type === 'style') {
            ctx.body = compileSFCStyle(cwd, descriptor.styles[Number(query.index)], query.index, filename, pathname);
            return;
        }
        // TODO custom blocks
    });
};
const parseCache = new Map();
async function parseSFC(cwd, filename, saveCache = false) {
    let content;
    try {
        content = await fs_1.promises.readFile(filename, 'utf-8');
    }
    catch (e) {
        return [];
    }
    const { descriptor, errors } = resolveVue_1.resolveCompiler(cwd).parse(content, {
        filename
    });
    if (errors) {
        // TODO
    }
    const prev = parseCache.get(filename);
    if (saveCache) {
        parseCache.set(filename, descriptor);
    }
    return [descriptor, prev];
}
exports.parseSFC = parseSFC;
function compileSFCMain(descriptor, pathname, timestamp) {
    timestamp = timestamp ? `&t=${timestamp}` : ``;
    // inject hmr client
    let code = `import "/__hmrClient"\n`;
    if (descriptor.script) {
        code += descriptor.script.content;
    }
    else {
        code += `export default {}`;
    }
    // The module rewriter will rewrite `export default {}` to
    // `let __script; export default (__script = {})
    let hasScoped = false;
    if (descriptor.styles) {
        descriptor.styles.forEach((s, i) => {
            if (s.scoped)
                hasScoped = true;
            code += `\nimport ${JSON.stringify(pathname + `?type=style&index=${i}${timestamp}`)}`;
        });
        if (hasScoped) {
            code += `\n__script.__scopeId = "data-v-${hash_sum_1.default(pathname)}"`;
        }
    }
    if (descriptor.template) {
        code += `\nimport { render as __render } from ${JSON.stringify(pathname + `?type=template${timestamp}`)}`;
        code += `\n__script.render = __render`;
    }
    code += `\n__script.__hmrId = ${JSON.stringify(pathname)}`;
    return code;
}
exports.compileSFCMain = compileSFCMain;
function compileSFCTemplate(cwd, template, filename, pathname, scoped) {
    const { code, errors } = resolveVue_1.resolveCompiler(cwd).compileTemplate({
        source: template.content,
        filename,
        compilerOptions: {
            scopeId: scoped ? `data-v-${hash_sum_1.default(pathname)}` : null,
            runtimeModuleName: '/__modules/vue'
        }
    });
    if (errors) {
        // TODO
    }
    return code;
}
exports.compileSFCTemplate = compileSFCTemplate;
function compileSFCStyle(cwd, style, index, filename, pathname) {
    const id = hash_sum_1.default(pathname);
    const { code, errors } = resolveVue_1.resolveCompiler(cwd).compileStyle({
        source: style.content,
        filename,
        id: `data-v-${id}`,
        scoped: style.scoped != null
    });
    // TODO css modules
    if (errors) {
        // TODO
    }
    return `
const id = "vue-style-${id}-${index}"
let style = document.getElementById(id)
if (!style) {
  style = document.createElement('style')
  style.id = id
  document.head.appendChild(style)
}
style.textContent = ${JSON.stringify(code)}
  `.trim();
}
exports.compileSFCStyle = compileSFCStyle;
